/*
 * Copyright (c) 2004 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-strcasecmp.c,v 1.3 2013/10/01 02:29:32 ca Exp $")

#include "sm/assert.h"
#include "sm/magic.h"
#include "sm/memops.h"
#include "sm/rpool.h"
#include "sm/test.h"
#include "sm/str.h"

#include <stdio.h>

#define SMAXLEN	256

static void
test_harness(sm_rpool_P a)
{
	sm_str_P s1, s2, s3, s4;
	size_t len;
	char *src1 = "abCdef";
	char *src2 = "abcDefg";
	char *src3 = "aBcdefh";
	char *src4 = "bcdef";

	len = strlen(src1) + strlen(src2);
	s1 = s2 = s3 = NULL;
	s1 = sm_str_scpy(a, src1, len);
	SM_TEST_E(s1 != NULL);
	SM_TEST(sm_str_getlen(s1) == strlen(src1));
	s2 = sm_str_scpy(a, src2, len);
	SM_TEST_E(s2 != NULL);
	SM_TEST(sm_str_getlen(s2) == strlen(src2));
	s3 = sm_str_scpy(a, src3, len);
	SM_TEST_E(s3 != NULL);
	SM_TEST(sm_str_getlen(s3) == strlen(src3));
	s4 = sm_str_scpy(a, src4, len);
	SM_TEST_E(s4 != NULL);
	SM_TEST(sm_str_getlen(s4) == strlen(src4));

	SM_TEST(sm_str_casecmp(s1, s2) < 0);
	SM_TEST(sm_str_casecmp(s1, s1) == 0);
	SM_TEST(sm_str_casecmp(s2, s1) > 0);

	SM_TEST(sm_str_casecmp(s1, s3) < 0);
	SM_TEST(sm_str_casecmp(s3, s3) == 0);
	SM_TEST(sm_str_casecmp(s3, s1) > 0);

	SM_TEST(sm_str_casecmp(s2, s3) < 0);
	SM_TEST(sm_str_casecmp(s2, s2) == 0);
	SM_TEST(sm_str_casecmp(s3, s2) > 0);

	SM_TEST(sm_str_casecmp_off(s1, 0, s2, 0) < 0);
	SM_TEST(sm_str_casecmp_off(s1, 0, s1, 0) == 0);
	SM_TEST(sm_str_casecmp_off(s2, 0, s1, 0) > 0);

	SM_TEST(sm_str_casecmp_off(s1, 0, s3, 0) < 0);
	SM_TEST(sm_str_casecmp_off(s3, 0, s3, 0) == 0);
	SM_TEST(sm_str_casecmp_off(s3, 0, s1, 0) > 0);

	SM_TEST(sm_str_casecmp_off(s2, 0, s3, 0) < 0);
	SM_TEST(sm_str_casecmp_off(s2, 0, s2, 0) == 0);
	SM_TEST(sm_str_casecmp_off(s3, 0, s2, 0) > 0);

	SM_TEST(sm_str_casecmp_off(s1, 0, s4, 0) < 0);
	SM_TEST(sm_str_casecmp_off(s1, 0, s4, 1) < 0);
	SM_TEST(sm_str_casecmp_off(s1, 0, s4, 2) < 0);
	SM_TEST(sm_str_casecmp_off(s1, 0, s4, 3) < 0);
	SM_TEST(sm_str_casecmp_off(s1, 0, s4, 4) < 0);
	SM_TEST(sm_str_casecmp_off(s1, 1, s4, 0) == 0);
	SM_TEST(sm_str_casecmp_off(s1, 2, s4, 1) == 0);
	SM_TEST(sm_str_casecmp_off(s1, 3, s4, 2) == 0);
	SM_TEST(sm_str_casecmp_off(s1, 4, s4, 3) == 0);
	SM_TEST(sm_str_casecmp_off(s1, 5, s4, 4) == 0);
	SM_TEST(sm_str_casecmp_off(s1, 6, s4, 5) == 0);

	sm_str_free(s1);
	sm_str_free(s2);
	sm_str_free(s3);
	sm_str_free(s4);
 error:
	return;
}

int
main(int argc, char *argv[])
{
	sm_rpool_P a;

	sm_test_begin(argc, argv, "test str_casecmp");

	/* create an rpool for entire test */
	a = sm_rpool_new(NULL);
	SM_TEST(a != NULL);
	test_harness(a);
	sm_rpool_delete(a);

	/* test without rpool */
	test_harness(NULL);
	return sm_test_end();
}
