#!/bin/sh
# $Id: t-tls-certpin-0.sh,v 1.20 2014/03/02 17:43:50 ca Exp $
#
# Copyright (c) 2014 Claus Assmann
#
# By using this file, you agree to the terms and conditions set
# forth in the license/LICENSE.3C file which can be found at the
# top level of this source code distribution.
#
# ----------------------------------------
# test TLS: cert pinning; check delivered mail
# uses smtps2 with TLS
# run smtpcnf.sh with different parameters
# ----------------------------------------
#

SD=`dirname $0`
if test "${SD}" = "."
then
   SD=../chkmts
fi
. ${SD}/common.sh

SINKPRG=${STE}/smtps3
chkcto smtpc MTA_USE_TLS
chkcto smtpc FFR_CERT_PINNING
sinkhastls
smtpctlsdebug

if test X"`uname`" != "XOSF1"
then
while getopts s:V FLAG
do
  case "${FLAG}" in
    s) SKIP="${SKIP} ${OPTARG}";;
    V) VERBOSE=true;;
  esac
done
shift `expr ${OPTIND} - 1`
fi

SRCPRG=${STE}/smtpc2
SENDER=sender@local.dom
RCPT=user@local.host

cd ../qmgr || exit 1

rm -f ${ALIMAP} ${MTMAP} ${ACCMAP} ${QMGRCONF} ${QMGRCONFMAP}
# create aliases
${CAT} > ${ALI} <<EOF
abuse:	local:
postmaster:	local:
user:	local:
sender: local:
EOF
# create map
${MM} < ${ALI}
${CAT} > ${ACC} <<EOF
cltaddr:${MTA_LOCALHOST}	relay
EOF
${MM} -t'	' -F ${ACCMAP} < ${ACC}

# XREF: smtpc/smclt.c
P="cert_fp does not match previous value"

${CAT} > ${SMXCNF} <<EOF
qmgr { Log_Level = 12; wait_for_server=4; wait_for_client=4;
  smtpc { flags = { lookup_rcpt_conf, lookup_session_conf, } }
}
smar { Log_Level = 12; nameserver = ${NS}; }
smtpc { Log_Level = 12; remote_port=${SNKPORT}; wait_for_server=4;
${TLSCDBG}
}
smtps {
Log_Level = 14;
daemon_address = localhost:${SRVPORT};
wait_for_server=4;
flags = {access};
}
EOF

# ----------------------------------------
# test: send mails to local address, check Received: header for TLS
# don't modify cert, so no error/log
TEST=1
if echo "${SKIP} X" | ${GREP} -v " ${TEST} " > /dev/null
then
ERR=false
if ${SHELL} ${SD}/smtpcnf.sh -r -Y "-@S -d9" -E 1 -A 1 -X 1 -S${SENDER} -R${RCPT} -z
then
  if chkesmtps ${S1}
  then
    if ${SRCPRG} -r ${H}:${SRVPORT} -f"${SENDER}" -R${RCPT} >> ${CLTL} 2>&1
    then
      chkesmtps ${S2}
      if fgrep "$P" ${SCL} >/dev/null
      then
        echo "$0: ${TEST}: found $P in ${SCL}"
        ERRS=`expr ${ERRS} + 1 `
      fi
    else
      echo "$0: ${TEST}: ${SRCPRG} failed"
      ERRS=`expr ${ERRS} + 1 `
    fi
  fi
else
  ERR=true
fi
${SD}/stopmts.sh
if ${ERR}
then
   echo "$0: test ${TEST} failed"
   ERRS=`expr ${ERRS} + 1 `
   test X"${MTA_STOPONERROR}" != X && exit 1
fi
fi


# ----------------------------------------
# test: send mails to local address, check Received: header for TLS
# change cert
TEST=2
if echo "${SKIP} X" | ${GREP} -v " ${TEST} " > /dev/null
then
ERR=false
if ${SHELL} ${SD}/smtpcnf.sh -r -Y "-@S -d4" -E 1 -A 1 -X 1 -S${SENDER} -R${RCPT} -z
then
  if chkesmtps ${S1}
  then

    # stop smtps3, change cert...
    SRVPID=`${TAIL} -n 1 ${PIDS}`
    if test X"${SRVPID}" != "X"
    then
      kill -TERM ${SRVPID}
    else
      echo "$0: SRVPID not found?"
    fi

# start SMTP sink
${SINKPRG} -C ./${SMTPDIR} -8 2 -@S -Ycsccert.pem -Yksckey.pem -d4 -l . -b ${H}:${SNKPORT} >> ${SRVL} 2>&1 &
status=$?
SRVPID=$!
if test "x${status}" != "x0"
then
  echo "$0: FAIL: cannot start server ${status}"
  exit ${status}
fi
sleep 1

    if ${SRCPRG} -r ${H}:${SRVPORT} -f"${SENDER}" -R${RCPT} >> ${CLTL} 2>&1
    then
      $PKILL -USR1 -U ${USER} qmgr
      chkesmtps ${S2}
      if fgrep "$P" ${SCL} >/dev/null
      then
        :
      else
        echo "$0: ${TEST}: missing $P in ${SCL}"
        ERRS=`expr ${ERRS} + 1 `
      fi
    else
      echo "$0: ${TEST}: ${SRCPRG} failed"
      ERRS=`expr ${ERRS} + 1 `
    fi

  fi
else
  ERR=true
fi
${SD}/stopmts.sh
kill ${SRVPID}
if ${ERR}
then
   echo "$0: test ${TEST} failed"
   ERRS=`expr ${ERRS} + 1 `
   test X"${MTA_STOPONERROR}" != X && exit 1
fi
fi


# ----------------------------------------
# test: cert_fp does not change;
# send to different addresses: cert_fp's must not be "mixed up".
TEST=3
if echo "${SKIP} X" | ${GREP} -v " ${TEST} " > /dev/null
then
ERR=false
if ${SHELL} ${SD}/smtpcnf.sh -r -Y "-@S -d4" -E 1 -A 1 -X 1 -S${SENDER} -R${RCPT} -z
then
  if chkesmtps ${S1}
  then

# start LMTP sink
${SINKPRG} -C ./${LMTPDIR} -d4 -L unix:lmtpsock >> ${SRVL} 2>&1 &
status=$?
SRVPID=$!
if test "x${status}" != "x0"
then
  echo "$0: FAIL: cannot start server ${status}"
  exit ${status}
fi
sleep 1

    ${SRCPRG} -r ${H}:${SRVPORT} -fs2${TEST}@local.host -Ruser@local.dom >> ${CLTL} 2>&1
    ${SRCPRG} -r ${H}:${SRVPORT} -fs3${TEST}@local.host -Ruser@local.host >> ${CLTL} 2>&1
    ${SRCPRG} -r ${H}:${SRVPORT} -fs4${TEST}@local.host -Ruser@local.host >> ${CLTL} 2>&1
    ${SRCPRG} -r ${H}:${SRVPORT} -fs5${TEST}@local.host -Ruser@local.dom >> ${CLTL} 2>&1

      $PKILL -USR1 -U ${USER} qmgr
      chkesmtps ${S2} || ERRS=`expr ${ERRS} + 1 `
      chkesmtps ${S3} || ERRS=`expr ${ERRS} + 1 `
      if fgrep "$P" ${SCL} >/dev/null
      then
        echo "$0: ${TEST}: found $P in ${SCL}"
        ERRS=`expr ${ERRS} + 1 `
      fi

  fi
else
  ERR=true
fi
${SD}/stopmts.sh
kill ${SRVPID}
if ${ERR}
then
   echo "$0: test ${TEST} failed"
   ERRS=`expr ${ERRS} + 1 `
   test X"${MTA_STOPONERROR}" != X && exit 1
fi
fi


# ----------------------------------------
# test: "wrong" cert_fp must not be sent to qmgr
# HMM??? need to define behaviour!!!
TEST=4
if echo "${SKIP} X" | ${GREP} -v " ${TEST} " > /dev/null
then

EDBROK="../qmgr/t-certpin-edbr.ok"
${CAT} > ${EDBROK} <<EOF
got transaction
	mail=<${SENDER}>
	rcpts_tot=1
	rcpts_left=1
	rcpts_temp=1
	rcpts_perm=0
	state=421
	aqt_rcpts_ar=0
got recipient
	rcpt=<${RCPT}>
	aqr_rcpt_idx=0
	aqr_da_idx=0
	aqr_status=421
EOF

# requirement can't be fulfilled
${CAT} > ${ACC} <<EOF
smtpc_rcpt_conf:${RCPT}	tls_requirements { flags={verified}; cipher_bits_min=1024;}
EOF
${MM} -t'	' -F ${ACCMAP} < ${ACC}

ERR=false
if ${SHELL} ${SD}/smtpcnf.sh -r -Y "-@S -d4" -E 0 -A 1 -X 1 -S${SENDER} -R${RCPT} -z -O ${EDBROK}
then
  ${SRCPRG} -r ${H}:${SRVPORT} -fs3${TEST}@local.host -Ruser@local.host >> ${CLTL} 2>&1

  $PKILL -USR1 -U ${USER} qmgr
  if fgrep "$P" ${SCL} >/dev/null
  then
    echo "$0: ${TEST}: found $P in ${SCL}"
    ERRS=`expr ${ERRS} + 1 `
  fi

else
  ERR=true
fi
${SD}/stopmts.sh
if ${ERR}
then
   echo "$0: test ${TEST} failed"
   ERRS=`expr ${ERRS} + 1 `
   test X"${MTA_STOPONERROR}" != X && exit 1
fi
fi


# ----------------------------------------
cd ${THISDIR}
rm -f ${MTMAP} ${ALIMAP}

testend
