#!/bin/sh
# $Id: t-tls-reqr-3.sh,v 1.13 2013/10/21 01:46:52 ca Exp $
#
# Copyright (c) 2013 Claus Assmann
#
# By using this file, you agree to the terms and conditions set
# forth in the license/LICENSE.3C file which can be found at the
# top level of this source code distribution.
#
# ----------------------------------------
# test TLS requirements in smtpc; check CN/subjectAltName
# creates two MTA (one in chkmts, one in qmgr) to use STARTTLS
# run smtpcnf.sh with different parameters
# ----------------------------------------
# need tests for:
# deferred mail (stored in defedb and tried again)
#
# tests need to be enhanced to properly check the results, i.e.,
# did really happen what was supposed to happen?
#
# CN=host.example.org
# DNS:mail.example.org, DNS:www.example.org, DNS:*.example.net
#

SD=`dirname $0`
if test "${SD}" = "."
then
   SD=../chkmts
fi
. ${SD}/common.sh

tlsrintro "$@"

setuptlsr0 "" "" "smtpc_rcpt_conf:user@local.dom	tls_requirements { flags={verified}; hostnames={ server.example.net} }" "tls { key_file = mtaaltkey.pem; cert_file=mtaaltcert.pem; }"

# ----------------------------------------
# test: send mail to sink, requirements fulfilled
# matches subjectAltName wildcard

TEST=1
if echo " ${SKIP} X" | ${GREP} -v " ${TEST} " > /dev/null
then
ERR=false

if ${SHELL} ${SD}/smtpcnf.sh -F ${SRVPORT} -3 lmtpsock -Ssender${TEST}@other.dom -Ruser@local.dom -E 0 -X 0 -A 1
then
  if test -s ${B1}
  then
    if ${GREP} 'TLS' ${B1} >/dev/null
    then
      :
    else
      ${VERBOSE} && echo "$0: test ${TEST}: missing TLS in ${B1}"
      ERR=true
    fi
  else
    ${VERBOSE} && echo "$0: test ${TEST}: missing ${B1}"
    ERR=true
  fi
else
  ERR=true
fi
if ${ERR}
then
   echo "$0: test ${TEST} failed"
   ERRS=`expr ${ERRS} + 1 `
   test X"${MTA_STOPONERROR}" != X && exit 1
fi
rm -f ${B1}
else
  ${VERBOSE} && echo "$0: test ${TEST} SKIPPED"
fi

# ----------------------------------------
# test: send mail to sink, requirements fulfilled
# multiple hostnames, matches subjectAltName

TEST=2
if echo " ${SKIP} X" | ${GREP} -v " ${TEST} " > /dev/null
then
ERR=false

cd ${THISDIR}
${CAT} > ${ACC} <<EOF
smtpc_rcpt_conf:user@local.dom	tls_requirements { flags={verified}; hostnames={ server.example.org, server.example.net } }
EOF
rm -f ${ACCMAP}
${MM} -t'	' -F ${ACCMAP} < ${ACC}
# notify smar
kill -USR1 ${SMARPID}
cd ../qmgr


if ${SHELL} ${SD}/smtpcnf.sh -F ${SRVPORT} -3 lmtpsock -Ssender${TEST}@other.dom -Ruser@local.dom -E 0 -X 0 -A 1
then
  if test -s ${B1}
  then
    if ${GREP} 'TLS' ${B1} >/dev/null
    then
      :
    else
      ${VERBOSE} && echo "$0: test ${TEST}: missing TLS in ${B1}"
      ERR=true
    fi
  else
    ${VERBOSE} && echo "$0: test ${TEST}: missing ${B1}"
    ERR=true
  fi
else
  ERR=true
fi
if ${ERR}
then
   echo "$0: test ${TEST} failed"
   ERRS=`expr ${ERRS} + 1 `
   test X"${MTA_STOPONERROR}" != X && exit 1
fi
rm -f ${B1}
else
  ${VERBOSE} && echo "$0: test ${TEST} SKIPPED"
fi

# ----------------------------------------
# test: send mail to sink, requirements fulfilled: 1-1 match with CN

TEST=3
if echo " ${SKIP} X" | ${GREP} -v " ${TEST} " > /dev/null
then
ERR=false

cd ${THISDIR}
${CAT} > ${ACC} <<EOF
smtpc_rcpt_conf:user@local.dom	tls_requirements { flags={verified}; hostnames={ host.example.org} }
EOF
rm -f ${ACCMAP}
${MM} -t'	' -F ${ACCMAP} < ${ACC}
# notify smar
kill -USR1 ${SMARPID}
cd ../qmgr


if ${SHELL} ${SD}/smtpcnf.sh -F ${SRVPORT} -3 lmtpsock -Ssender${TEST}@other.dom -Ruser@local.dom -E 0 -X 0 -A 1
then
  if test -s ${B1}
  then
    if ${GREP} 'TLS' ${B1} >/dev/null
    then
      :
    else
      ${VERBOSE} && echo "$0: test ${TEST}: missing TLS in ${B1}"
      ERR=true
    fi
  else
    ${VERBOSE} && echo "$0: test ${TEST}: missing ${B1}"
    ERR=true
  fi
else
  ERR=true
fi
if ${ERR}
then
   echo "$0: test ${TEST} failed"
   ERRS=`expr ${ERRS} + 1 `
   test X"${MTA_STOPONERROR}" != X && exit 1
fi
rm -f ${B1}
else
  ${VERBOSE} && echo "$0: test ${TEST} SKIPPED"
fi



# ----------------------------------------
# test: send mail to sink, requirements not fulfilled

TEST=4
if echo " ${SKIP} X" | ${GREP} -v " ${TEST} " > /dev/null
then
ERR=false

cd ${THISDIR}
${CAT} > ${ACC} <<EOF
smtpc_rcpt_conf:user@local.dom	tls_requirements { flags={verified}; hostnames={ my.server.example.org} }
EOF
rm -f ${ACCMAP}
${MM} -t'	' -F ${ACCMAP} < ${ACC}
# notify smar
kill -USR1 ${SMARPID}
cd ../qmgr

if ${SHELL} ${SD}/smtpcnf.sh -F ${SRVPORT} -3 lmtpsock -Ssender${TEST}@other.dom -Ruser@local.dom -E 0 -X 0 -A 1
then
  if test -s ${B1}
  then
    ${VERBOSE} && echo "$0: test ${TEST}: found ${B1}"
    ERR=true
  else
    # check that mail is in queue, check error message...
    # how?
    cd ${THISDIR} || ERR=true
    ls ?/S* >/dev/null || ERR=true
# cannot use edbr as qmgr is still running
#     ${EDBR} -V > ${OUT}
#     if ${GREP} 'text=421 4.7.0 smtpc security requirements not met' ${OUT} >/dev/null
#     then
#       :
#     else
#       ${VERBOSE} && echo "$0: test ${TEST}: missing 421 text"
#       ERR=true
#     fi
    cd ../qmgr
  fi
else
  ERR=true
fi
if ${ERR}
then
   echo "$0: test ${TEST} failed"
   ERRS=`expr ${ERRS} + 1 `
   test X"${MTA_STOPONERROR}" != X && exit 1
fi
rm -f ${B1}
else
  ${VERBOSE} && echo "$0: test ${TEST} SKIPPED"
fi


testend rmb
