/*
 * $Id: sm-expr.h,v 1.18 2010/06/11 02:31:59 ca Exp $
 */

#ifndef SM_EXPR_H
#define SM_EXPR_H 1

/* lexer/parser context */
struct lpctx_S {
	const char	*lpc_text;
	int			 lpc_text_n;
	int			 lpc_pos;

	cexpr_P		 lpc_expr;
	uint		 lpc_rooti;

	/* symbol table for currently valid ids? how to pass in? */
	sm_id_entry_P	lpc_idtab;
	size_t			lpc_idsz;
};
typedef struct lpctx_S lpctx_T, *lpctx_P;

/* function table */
struct sm_func_S
{
	const char	*f_id;
	uint		 f_len;
	int			 f_opcode;
	/* others? e.g., nargs, signature, ...? */
};
typedef struct sm_func_S	sm_func_T, *sm_func_P;

#define SMBSRCHN(pfx, idtabT, idef, lenf)		\
int	\
pfx##_bsrchn(idtabT idtab, size_t sz, const char *id, int len)	\
{	\
	int cmp, low, mid, high;	\
	\
	assert(idtab != NULL);	\
	assert(id != NULL);	\
	assert(sz > 0);	\
	\
	low = 0;	\
	high = sz - 1;	\
	while (low <= high) {	\
		mid = (low + high) / 2;	\
		cmp = strncmp(id, idtab[mid].idef, len);	\
		if (cmp == 0)	\
			cmp = len - idtab[mid].lenf;	\
		if (cmp == 0 && len == idtab[mid].lenf)	\
			return mid;	\
		else if (cmp < 0)	\
			high = mid - 1;	\
		else if (cmp > 0)	\
			low = mid + 1;	\
	}	\
	return -1;	\
}

#define SMBINSRCHN(pfx, idtab, sz, idef, lenf)		\
int	\
pfx##_binsrchn(const char *id, int len)	\
{	\
	int cmp, low, mid, high;	\
	\
	assert(id != NULL);	\
	assert(idtab != NULL);	\
	assert(sz > 0);	\
	\
	low = 0;	\
	high = sz - 1;	\
	while (low <= high) {	\
		mid = (low + high) / 2;	\
		cmp = strncmp(id, idtab[mid].idef, len);	\
		if (cmp == 0)	\
			cmp = len - idtab[mid].lenf;	\
		if (cmp == 0 && len == idtab[mid].lenf)	\
			return mid;	\
		else if (cmp < 0)	\
			high = mid - 1;	\
		else if (cmp > 0)	\
			low = mid + 1;	\
	}	\
	return -1;	\
}

sm_ret_T m_expr_init(uint _n, cexpr_P *_pexpr);
sm_ret_T m_expr_free(cexpr_P _expr);
sm_ret_T m_expr_get_node(cexpr_P _expr, uint *_pexi);
int crt_exp(lpctx_P _lpctx, int _nargs, op_T _op, uint const *_args);
int crt_id(lpctx_P _lpctx, id_P _id, val_type_T _vt);
int crt_str(lpctx_P _lpctx, id_P _id);
int crt_int(lpctx_P _lpctx, int _v);
int crt_exp_1(lpctx_P _lpctx, op_T _op, uint _right);
int crt_exp_2(lpctx_P _lpctx, op_T _op, uint _left, uint _right);
int crt_exp_3(lpctx_P _lpctx, op_T _op, uint _bexp, uint _tbranch, uint _fbranch);

int e_bsrchn(sm_id_entry_P idtab, size_t _sz, const char *_id, int _len);
int f_binsrchn(const char *_id, int _len);

int yyparse(void *_ctx);

#endif /* SM_EXPR_H */
