/*
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-evthr-p0.c,v 1.4 2011/02/02 21:30:04 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/memops.h"
#include "sm/heap.h"
#include "sm/test.h"
#include "sm/evthr.h"
#include "sm/io.h"
#include "sm/fxszq.h"

#include <stdio.h>

/*
*/

struct fxszq_S
{
	uint	 fsq_max;   /* allocated size */
	uint	 fsq_cur;   /* currently used (basically last-first) */
	uint	 fsq_first; /* first index to read */
	uint	 fsq_last;  /* last index to read (first to write) */
	int	*fsq_array; /* array */
};
typedef struct fxszq_S	fxszq_T, *fxszq_P;
fxszq_T q1;

#define WHAT_TERM	0
#define WHAT_CONT	1

#define REPS 5

#define T_QS	256

static int Verbose = 0;
static long reps = 0;
static sm_evthr_task_P	task, task2;
static bool stop = false;

struct t_ctx_S
{
	sm_evthr_ctx_P	 t_ctx;
	int				 t_what;
	int				 t_called;
	sm_evthr_task_P	 t_wakeup;
};
typedef struct t_ctx_S t_ctx_T, *t_ctx_P;

pthread_mutex_t	 q_mutex;
pthread_cond_t	 q_cond;
long q_add, q_rm;

static void
t_init(uint sz)
{
	int r;

	SM_TEST(sz != 0);
	if (sz == 0)
		return;
	q1.fsq_array = sm_malloc(sizeof(int) * sz);
	SM_TEST(q1.fsq_array != NULL);
	if (q1.fsq_array == NULL)
		return;
	FSQ_INIT(sz, q1.fsq_max, q1.fsq_cur, q1.fsq_first, q1.fsq_last);
	SM_TEST(FSQ_IS_EMPTY(q1.fsq_cur));
	SM_TEST(!FSQ_IS_FULL(q1.fsq_max, q1.fsq_cur));
	r = pthread_mutex_init(&q_mutex, NULL);
	r = pthread_cond_init(&q_cond, NULL);
}

static bool
add_q(int i)
{
	int r;
	bool f;

	r = pthread_mutex_lock(&q_mutex);
	if (FSQ_IS_FULL(q1.fsq_max, q1.fsq_cur)) {
		r = pthread_mutex_unlock(&q_mutex);
		return true;
	}
	FSQ_APPEND(q1.fsq_max, q1.fsq_last, q1.fsq_cur, q1.fsq_array, i);
	SM_TEST(!FSQ_IS_EMPTY(q1.fsq_cur));
	++q_add;
	f = FSQ_IS_FULL(q1.fsq_max, q1.fsq_cur);
	r = pthread_mutex_unlock(&q_mutex);
	return f;
}

static bool
rm_q(void)
{
	int j, h, r;
	bool e;

	r = pthread_mutex_lock(&q_mutex);
	if (FSQ_IS_EMPTY(q1.fsq_cur)) {
		r = pthread_mutex_unlock(&q_mutex);
		return true;
	}
	j = FSQ_GET(q1.fsq_max, q1.fsq_first, q1.fsq_cur, q1.fsq_array, h);
	++q_rm;
	e = FSQ_IS_EMPTY(q1.fsq_cur);
	r = pthread_mutex_unlock(&q_mutex);
	return e;
}

static sm_ret_T
fct1(sm_evthr_task_P tsk)
{
	int i;
	bool f;
	t_ctx_P fctx;

	if (stop)
		return EVTHR_TERM|EVTHR_DEL;
	SM_ASSERT(tsk != NULL);
	fctx = (t_ctx_P) tsk->evthr_t_actx;
	if (fctx->t_called++ > reps)
		return EVTHR_TERM|EVTHR_DEL;

	for (i = 0; i < 10 && !(f = add_q(i)); i++)
		;
	if (f && !stop)
		i = evthr_wakeup_task(fctx->t_wakeup);

	switch (fctx->t_what) {
	  case WHAT_TERM:
		return EVTHR_TERM;
	  case WHAT_CONT:
	  default:
		return EVTHR_WAITQ;
	}
	/* NOTREACHED */
	return EVTHR_TERM|EVTHR_DEL;
}

static sm_ret_T
fct2(sm_evthr_task_P tsk)
{
	int i;
	bool e;
	t_ctx_P fctx;
	timeval_T sleept, delay, wakeup;

	if (stop)
		return EVTHR_TERM|EVTHR_DEL;
	SM_ASSERT(tsk != NULL);
	fctx = (t_ctx_P) tsk->evthr_t_actx;
	i = gettimeofday(&sleept, NULL);
	delay.tv_sec = 0;
	delay.tv_usec = 1;
	timeradd(&sleept, &delay, &wakeup);

	for (i = 0; i < 5 && !(e = rm_q()); i++)
		;

	if (e && !stop)
		i = evthr_wakeup_task(fctx->t_wakeup);

	if (++fctx->t_called > reps)
		return EVTHR_TERM|EVTHR_DEL;
	timeradd(&sleept, &delay, &tsk->evthr_t_sleep);
	return EVTHR_SLPQ;
}

static void
testev(int what, uint sz)
{
	int r;
	sm_ret_T ret;
	sm_evthr_ctx_P evthr_ctx;
	t_ctx_T tctx, tctx2;
	timeval_T sleept;

	t_init(sz);
	ret = thr_init();
	SM_TEST(sm_is_success(ret));
	sm_memzero(&sleept, sizeof(sleept));
	ret = evthr_init(&evthr_ctx, 1, 6, 10);
	SM_TEST(sm_is_success(ret));
	SM_TEST(evthr_ctx != NULL);

	tctx.t_called = 0;
	tctx.t_ctx = evthr_ctx;
	tctx.t_what = what;
	ret = evthr_task_new(evthr_ctx, &task, EVTHR_EV_SL, -1,
			&sleept, fct1, (void *) &tctx);
	SM_TEST(sm_is_success(ret));
	SM_TEST(task != NULL);

	r = gettimeofday(&sleept, NULL);
	SM_TEST(r == 0);
	sleept.tv_sec += 1;
	tctx2.t_called = 0;
	ret = evthr_task_new(evthr_ctx, &task2, EVTHR_EV_SL, -1,
			&sleept, fct2, (void *) &tctx2);
	SM_TEST(sm_is_success(ret));
	SM_TEST(task2 != NULL);

	tctx.t_wakeup = task2;
	tctx2.t_wakeup = task;

	ret = evthr_loop(evthr_ctx);
	SM_TEST(sm_is_success(ret));
	if (!sm_is_success(ret))
		fprintf(stderr, "evthr_loop()=%x\n", ret);

	/*
	**  we should "hold" the system before deleting tasks?
	**  deleting the tasks while they are still in use
	**  will break things.
	*/

	stop = true;
	SM_TEST(tctx.t_called > 0);
	if (Verbose > 0)
		fprintf(stderr, "fcts=%d\n", tctx.t_called);
	fprintf(stderr, "adds=%ld, rms=%ld\n", q_add, q_rm);

	ret = evthr_stop(evthr_ctx, 0);
	SM_TEST(sm_is_success(ret));
	if (!sm_is_success(ret))
		fprintf(stderr, "evthr_stop()=%x\n", ret);
	ret = thr_stop();
	SM_TEST(sm_is_success(ret));
}

static void
usage(const char *prg)
{
	fprintf(stderr, "usage: %s [options] socket\n", prg);
	fprintf(stderr, "options:\n");
	fprintf(stderr, "-l n    number of loops\n");
	fprintf(stderr, "-r n    number of repetitions\n");
	fprintf(stderr, "-w n    value for 'what'\n");
	fprintf(stderr, "-V      increase verbosity\n");
}

int
main(int argc, char *argv[])
{
	int c, what, loops;
	uint sz;

	what = 1;
	loops = 16;
	reps = 1;
	sz = 100;
	while ((c = getopt(argc, argv, "l:r:w:V")) != -1) {
		switch (c) {
		  case 'l':
			loops = atoi(optarg);
			break;
		  case 'r':
			reps = atoi(optarg);
			break;
		  case 'w':
			what = atoi(optarg);
			break;
		  case 'V':
			Verbose++;
			break;
		  default:
			usage(argv[0]);
			return(1);
		}
	}
	sm_test_begin(argc, argv, "test evthr perf 0");
	testev(what, sz);
	return sm_test_end();
}
