/*
 * Copyright (c) 2005 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: hdrmodins.c,v 1.6 2014/01/30 16:37:29 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/hdrmod.h"

/*
**  SM_HDRMOD_INSERT -- insert hdrmod entry at right place in list
**
**	Parameters:
**		sm_hdrmodhd -- head of header modification list
**		sm_hdrmod -- header modification entry to insert
**
**	Returns:
**		SM_SUCCESS
*/

static int hdrmod_weight[] = {
	0, /* SM_HDRMOD_T_PREPEND */
	1, /* SM_HDRMOD_T_INSERT */
	1, /* SM_HDRMOD_T_REPLACE */
	1, /* SM_HDRMOD_T_REMOVE */
	2, /* SM_HDRMOD_T_APPEND */
};

sm_ret_T
sm_hdrmod_insert(sm_hdrmodhd_P sm_hdrmodhd, sm_hdrmod_P sm_hdrmod)
{
	SM_ASSERT(sm_hdrmodhd != NULL);
	SM_ASSERT(sm_hdrmod != NULL);
	if (!HDRMODL_EMPTY(sm_hdrmodhd)) {
		sm_hdrmod_P sm_hdrmod_h;

		for (sm_hdrmod_h = HDRMODL_FIRST(sm_hdrmodhd);
		     sm_hdrmod_h != HDRMODL_END(sm_hdrmodhd);
		     sm_hdrmod_h = HDRMODL_NEXT(sm_hdrmod_h))
		{
			if (hdrmod_weight[sm_hdrmod->sm_hm_type]
			    < hdrmod_weight[sm_hdrmod_h->sm_hm_type] ||
			    (SM_HM_USE_POS(sm_hdrmod->sm_hm_type) &&
			     SM_HM_USE_POS(sm_hdrmod_h->sm_hm_type) &&
			     sm_hdrmod->sm_hm_pos < sm_hdrmod_h->sm_hm_pos))
			{
				HDRMODL_INS_BEFORE(sm_hdrmodhd, sm_hdrmod, sm_hdrmod_h);
				return SM_SUCCESS;
			}
		}
	}
	HDRMODL_APP(sm_hdrmodhd, sm_hdrmod);
	return SM_SUCCESS;
}
