/* included by rcbputv.c and rcbputrec.c */
	for (;;) {
		k = va_arg(ap, int);
		if (SM_RCBV_END == k)
			break;
		rt = va_arg(ap, uint32_t);
		switch (k) {
		  case SM_RCBV_INT:
			v = va_arg(ap, uint32_t);
			if (RT_NOSEND == rt)
				break;
			ret = sm_rcb_put3uint32(rcb, 4, rt, v);
			if (sm_is_err(ret))
				goto err1;
			break;
		  case SM_RCBV_INT64:
			{
				uint64_t u;

				u = va_arg(ap, uint64_t);
				if (RT_NOSEND == rt)
					break;
				ret = sm_rcb_put3uint64(rcb, 8, rt, u);
				if (sm_is_err(ret))
					goto err1;
			}
			break;
#if MTA_USE_RCBV_INT2
		  case SM_RCBV_INT2:
			{
				uint32_t v2;

				v = va_arg(ap, uint32_t);
				v2 = va_arg(ap, uint32_t);
				if (RT_NOSEND == rt)
					break;
				ret = sm_rcb_put4uint32(rcb, 8, rt, v, v2);
				if (sm_is_err(ret))
					goto err1;
			}
			break;
#endif /* MTA_USE_RCBV_INT2 */
		  case SM_RCBV_STR:
			str = va_arg(ap, sm_str_P);
			if (RT_NOSEND == rt)
				break;
			if (NULL == str) {
				ret = sm_rcb_put2uint32(rcb, 0, rt);
				if (sm_is_err(ret))
					goto err1;
				break;
			}
			SM_IS_BUF(str);
			ret = sm_rcb_put2uint32(rcb, sm_str_getlen(str), rt);
			if (sm_is_err(ret))
				goto err1;
			ret = sm_rcb_putn(rcb, sm_str_data(str), sm_str_getlen(str));
			if (sm_is_err(ret))
				goto err1;
			break;

#if MTA_USE_RCBV_STRN
		  case SM_RCBV_STRA:
			{
				uint nv;
				sm_str_P *strv;

				nv = va_arg(ap, uint);
				strv = va_arg(ap, sm_str_P *);
				if (RT_NOSEND == rt)
					break;
				ret = sm_rcb_putastr(rcb, rt, nv, strv);
				if (sm_is_err(ret))
					goto err1;
			}
			break;
#endif /* MTA_USE_RCBV_STRN */

#if !SM_NO_CSTR
		  case SM_RCBV_CSTR:
			cstr = va_arg(ap, sm_cstr_P);
			if (RT_NOSEND == rt)
				break;
			if (NULL == cstr) {
				ret = sm_rcb_put2uint32(rcb, 0, rt);
				if (sm_is_err(ret))
					goto err1;
				break;
			}
			SM_IS_CSTR(cstr);
			ret = sm_rcb_put2uint32(rcb, sm_cstr_getlen(cstr), rt);
			if (sm_is_err(ret))
				goto err1;
			ret = sm_rcb_putn(rcb, sm_cstr_data(cstr), sm_cstr_getlen(cstr));
			if (sm_is_err(ret))
				goto err1;
			break;
#endif /* !SM_NO_CSTR */
		  case SM_RCBV_BUF:
			buf = va_arg(ap, uchar *);
			len = va_arg(ap, uint);
			if (RT_NOSEND == rt)
				break;
			if (NULL == buf) {
				/* len should be 0 */
				ret = sm_rcb_put2uint32(rcb, 0, rt);
				if (sm_is_err(ret))
					goto err1;
				break;
			}
			ret = sm_rcb_put2uint32(rcb, len, rt);
			if (sm_is_err(ret))
				goto err1;
			ret = sm_rcb_putn(rcb, buf, len);
			if (sm_is_err(ret))
				goto err1;
			break;
#if MTA_USE_RCBV_INTN
		  case SM_RCBV_INTA:
			{
				uint32_t *av;
				uint nv;

				nv = va_arg(ap, uint);
				av = va_arg(ap, uint32_t *);
				if (RT_NOSEND == rt)
					break;
				ret = sm_rcb_putauint32(rcb, rt, nv, av);
				if (sm_is_err(ret))
					goto err1;
			}
			break;
#endif /* MTA_USE_RCBV_INTN */
		  case SM_RCBV_SOCK:
			sockaddr = va_arg(ap, sm_sockaddr_P);
			if (RT_NOSEND == rt || sockaddr == NULL)
				break;
			/* family, port, addr? */
			switch (sockaddr->sa.sa_family) {
			  case AF_INET:
				len = sizeof(sockaddr->sin.sin_addr);
				buf = (uchar *)&sockaddr->sin.sin_addr;
				k = sockaddr->sin.sin_port;
				break;

#if MTA_NETINET6
			  case AF_INET6:
				len = sizeof(sockaddr->sin6.sin6_addr);
				buf = (uchar *)&sockaddr->sin6.sin6_addr;
				k = sockaddr->sin6.sin6_port;
				break;
#endif
			  default:
				ret = sm_error_perm(SM_EM_NONE, EPROTONOSUPPORT);
				goto err1;
			}
			ret = sm_rcb_put4uint32(rcb, len + 8, rt, sockaddr->sa.sa_family, k);
			if (sm_is_err(ret))
				goto err1;
			ret = sm_rcb_putn(rcb, buf, len);
			if (sm_is_err(ret))
				goto err1;
			break;

		  case SM_RCBV_IPADDR:
			ipaddr = va_arg(ap, sm_ipaddr_P);
			if (RT_NOSEND == rt || ipaddr == NULL)
				break;
			/* family, port, addr? */
			switch (ipaddr->ipa_family) {
			  case AF_INET:
				len = sizeof(ipaddr->ipa_addr.ipv4_addr);
				buf = (uchar *)&ipaddr->ipa_addr.ipv4_addr;
				break;

#if MTA_NETINET6
			  case AF_INET6:
				len = sizeof(ipaddr->ipa_addr.ipv6_addr);
				buf = (uchar *)&ipaddr->ipa_addr.ipv6_addr;
				break;
#endif
			  default:
				ret = sm_error_perm(SM_EM_NONE, EPROTONOSUPPORT);
				goto err1;
			}
			ret = sm_rcb_put3uint32(rcb, len + 4, rt, ipaddr->ipa_family);
			if (sm_is_err(ret))
				goto err1;
			ret = sm_rcb_putn(rcb, buf, len);
			if (sm_is_err(ret))
				goto err1;
			break;
		  default:
			ret = sm_error_perm(SM_EM_RECCOM, EINVAL);
			goto err1;
		}
	}
