/*
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-err2ex-0.c,v 1.3 2013/12/17 15:31:18 ca Exp $")

#include "sm/error.h"
#include "sm/wait.h"
#include "sm/sysexits.h"
#include "sm/io.h"
#include "sm/test.h"

static sm_ret_T
t_exv(int exv)
{
	int r;
	sm_error_T err;

	err = sm_ex_error(0, exv);
	SM_TEST(sm_is_ex_error(err));
	r = err2ex(err);
	SM_TEST(r == exv);
	return r;
}

static sm_ret_T
t_err2ex(void)
{
	int r;
	sm_error_T err;

	err = 0;
	r = err2ex(err);
	SM_TEST(EX_OK == r);

	err = sm_err_temp(ECONNREFUSED);
	r = err2ex(err);
	SM_TEST(EX_TEMPFAIL == r);

	err = sm_err_temp(ENOMEM);
	r = err2ex(err);
	SM_TEST(EX_TEMPFAIL == r);

	err = sm_err_perm(EIO);
	r = err2ex(err);
	SM_TEST(EX_UNAVAILABLE == r);

	t_exv(EX_CONFIG);
	t_exv(EX_USAGE);
	t_exv(EX_TEMPFAIL);
	t_exv(EX_OSERR);
	t_exv(EX_DATAERR);

	return r;
}

extern int optind;

int
main(int argc, char *argv[])
{
	int c;
	bool interactive;

	interactive = false;
	while ((c = getopt(argc, argv, "i")) != -1) {
		switch (c) {
		  case 'i':
			interactive = true;
			break;
		  default:
			return EX_USAGE;
		}
	}
	sm_test_begin(argc, argv, "test err2ex");
	argc -= optind;
	argv += optind;

	if (interactive) {
		int i, r;

		for (i = 0; i < argc; i++) {
			c = (int) strtol(argv[i], (char **)NULL, 0);
			r = err2ex(c);
			sm_io_fprintf(smioerr, "status=%x, ex=%d\n", c, r);
		}
	}
	else
		t_err2ex();
	return sm_test_end();
}
