/*
 *  consumer program trying to emulate smtps receiver side in qmgr
 *  (not very closely however).
 *  uses pthreads directly.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-consumer-0.c,v 1.16 2010/09/08 02:17:51 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/memops.h"
#include "sm/heap.h"
/* #include "sm/test.h" */
#include "sm/fxszq.h"
#include "sm/socket.h"
#include "sm/unixsock.h"
#include "sm/io.h"
#include "sm/misc.h"

#include <pthread.h>

static pthread_mutex_t	 q_mutex;
static pthread_cond_t	 q_cond;
static pthread_t task1, task2;
static int fd;
static long q_add, q_rm;

#define LOCK_Q r = pthread_mutex_lock(&q_mutex)
#define UNLOCK_Q r = pthread_mutex_unlock(&q_mutex);

static sm_ret_T qm_listen(const char *_sockname, int *_pfd);

struct fxszq_S
{
	uint	 fsq_max;   /* allocated size */
	uint	 fsq_cur;   /* currently used (basically last-first) */
	uint	 fsq_first; /* first index to read */
	uint	 fsq_last;  /* last index to read (first to write) */
	int	*fsq_array; /* array */
};
typedef struct fxszq_S	fxszq_T, *fxszq_P;
fxszq_T q1;

#include "t-prdcns-0.h"

struct t_ctx_S
{
	int			 t_called;
	int			 t_once;
	int			 t_max;
	int			 t_buf[T_QS];
};
typedef struct t_ctx_S t_ctx_T, *t_ctx_P;

static int
t_init(uint sz)
{
	int r;
	sm_ret_T ret;

	SM_TEST(sz != 0);
	if (sz == 0)
		return -1;
	q1.fsq_array = sm_malloc(sizeof(int) * sz);
	SM_TEST(q1.fsq_array != NULL);
	if (q1.fsq_array == NULL)
		return -1;
	FSQ_INIT(sz, q1.fsq_max, q1.fsq_cur, q1.fsq_first, q1.fsq_last);
	SM_TEST(FSQ_IS_EMPTY(q1.fsq_cur));
	SM_TEST(!FSQ_IS_FULL(q1.fsq_max, q1.fsq_cur));
	r = pthread_mutex_init(&q_mutex, NULL);
	SM_TEST(r == 0);
	r = pthread_cond_init(&q_cond, NULL);
	SM_TEST(r == 0);
	ret = qm_listen(MYSOCK, &fd);
	return ret;
}

static int
add_q(int i)
{
	int r, p;

	LOCK_Q;
	if (FSQ_IS_FULL(q1.fsq_max, q1.fsq_cur)) {
		UNLOCK_Q;
		return 100;
	}
	FSQ_APPEND(q1.fsq_max, q1.fsq_last, q1.fsq_cur, q1.fsq_array, i);
	SM_TEST(!FSQ_IS_EMPTY(q1.fsq_cur));
	++q_add;
	if (q_add > reps) stop = true;
	p = FSQ_PERC(q1.fsq_max, q1.fsq_cur);
	UNLOCK_Q;
	return p;
}

static int
perc_q(void)
{
	int r, p;

	LOCK_Q;
	p = FSQ_PERC(q1.fsq_max, q1.fsq_cur);
	UNLOCK_Q;
	return p;
}

static int
rm_q(int *pfree, int *pused, int *pv, int *pnelems)
{
	int i, h, r, n;

	LOCK_Q;
	if (FSQ_IS_EMPTY(q1.fsq_cur)) {
		UNLOCK_Q;
		return 0;
	}
	n = SM_MIN(q1.fsq_cur, *pnelems);
	for (i = 0; i < n; i++) {
		pv[i] = FSQ_GET(q1.fsq_max, q1.fsq_first, q1.fsq_cur, q1.fsq_array, h);
		++q_rm;
	}
	*pnelems = i;
	h = FSQ_PERC(q1.fsq_max, q1.fsq_cur);
	*pfree = q1.fsq_max - q1.fsq_cur;
	*pused = q1.fsq_cur;
	UNLOCK_Q;
	return h;
}

/* accept a single connection */
static sm_ret_T
qm_listen(const char *sockname, int *pfd)
{
	int lfd, r;
	sm_ret_T ret;
	struct sockaddr addr;
	sockaddr_len_T addrlen;

	SM_REQUIRE(sockname != NULL);
	SM_REQUIRE(pfd != NULL);
	ret = SM_SUCCESS;
	r = unlink(sockname);
	if (r < 0 && errno != ENOENT) {
		ret = sm_err_perm(errno);
		dbg_print(
			"sev=ERROR, func=qm_listen, socket=%s, unlink=%m\n",
			sockname, ret);
		goto error;
	}
	lfd = unix_server_listen(sockname, 10);
	if (!is_valid_socket(lfd)) {
		ret = sm_err_perm(errno);
		dbg_print(
			"sev=ERROR, func=qm_listen, socket=%s, unix_server_listen=%m\n",
			sockname, lfd);
		goto error;
	}

	addrlen = sizeof(addr);
	if (Verbose > 1)
		dbg_print("srv: accept\n");
	fd = unix_server_accept(lfd, &addr, &addrlen);
	SM_TEST(fd >= 0);
	if (fd < 0) {
		ret = sm_err_perm(errno);
		goto error;
	}
	if (Verbose > 1)
		dbg_print("srv: accepted fd=%d\n", fd);
	*pfd = fd;

	/* FALLTHROUGH */
  error:
	return ret;
}

static int
sndv(int32_t v)
{
	int p;
	char buf[PBUFSZ];

	*((int *)buf) = v;
	p = write(fd, buf, PBUFSZ);
	if (p <= 0)
		return sm_err_perm(errno);
	++q_sent;
	return p;
}

static int
rcvv(int32_t *pv)
{
	int r;
	char buf[PBUFSZ];

	r = read(fd, buf, sizeof(buf));
	if (r <= 0)
		return sm_err_perm(errno);
	++q_rcvd;
	*pv = *((int *)buf);
	return r;
}

static void *
t_rcv(void *ctx)
{
	t_ctx_P fctx;

	SM_ASSERT(ctx != NULL);
	fctx = (t_ctx_P) ctx;

	while (1) {
		int i, r, p;

		r = 0;
		if (Verbose > 0)
			dbg_print("t_rcv: c=%d\n", fctx->t_called);
		fctx->t_called++;
		if (stop) return 0;

		for (i = 0; i < fctx->t_once; i++) {
			int32_t v;

			if (Verbose > 2)
				dbg_print("t_rcv: i=%d, where=before_rcvv\n", i);
			r = rcvv(&v);
			if (Verbose > 2)
				dbg_print("t_rcv: i=%d, rcvv=%m, v=%d\n", i, r, v);
			if (r <= 0)
				goto error;
			p = add_q(v);
			if (Verbose > 2)
				dbg_print("t_rcv: i=%d, add_q=%d\n", i, p);
			if (p > 25) { /* maximum percentage? */
				int h;

				h = pthread_cond_signal(&q_cond);
				if (h != 0 || Verbose > 2)
					dbg_print("t_rcv: i=%d, signal=%d\n", i, h);
				do {
					p = perc_q();
					if (p > 75) {
						h = sm_us_sleep(100);
						if (h != 0 || Verbose > 2)
							dbg_print("t_rcv: i=%d, p=%d, sleep=%d\n", i, p, h);
					}
					if (stop) return 0;
				} while (p > 75);
				break;
			}
		}
		if (Verbose > 0)
			dbg_print("t_rcv: r=%d, stop=%d\n", r, stop);

		if (stop) return 0;
	}

 error:
	return 0;
}

static void *
t_ack(void *ctx)
{
	int i, r, n;
	int elems[T_QS];
	t_ctx_P fctx;

	SM_ASSERT(ctx != NULL);
	fctx = (t_ctx_P) ctx;

	r = 0;
	while (1) {
		int v, p, free, used;

		if (Verbose > 0)
			dbg_print("	t_ack: c=%d\n", fctx->t_called);
		fctx->t_called++;
		if (stop) return 0;

		i = pthread_mutex_lock(&q_mutex);
		SM_TEST_E(i == 0);
		if (Verbose > 2)
			dbg_print("	t_ack: where=afterlock, stat=%d, stop=%d\n", i, stop);
		while (!stop) {
			struct timespec w;
			struct timeval now;

			gettimeofday(&now, NULL);
			w.tv_sec = now.tv_sec + 5;
			w.tv_nsec = 0;
			if (Verbose > 2)
				dbg_print("	t_ack: where=beforewait\n", i);
			i = pthread_cond_timedwait(&q_cond, &q_mutex, &w);
			if (i != 0) {
				dbg_print("	t_ack: where=afterwait, stat==%d\n", i);
				if (ETIMEDOUT == i)
					goto error;
			}
			SM_TEST_E(i == 0);
			if (Verbose > 0)
				dbg_print("	t_ack: afterwait=%d, empty=%d\n", i,
					FSQ_IS_EMPTY(q1.fsq_cur));
			i = pthread_mutex_unlock(&q_mutex);
			SM_TEST_E(i == 0);
			if (Verbose > 0)
				dbg_print("	t_ack: afterlock=%d\n", i);
			if (stop) return 0;
			break;
		}

		do {
			/* get several elements from queue */
			n = SM_MIN(fctx->t_once, T_QS);
			if (Verbose > 2)
				dbg_print("	t_ack: n=%d, before=rm_q\n", n);
			p = rm_q(&free, &used, elems, &n);
			if (n == 0 && Verbose > 0 || Verbose > 2)
				dbg_print("	t_ack: n=%d, free=%d, used=%d\n", n, free, used);

			for (i = 0; i < n; i++) {
				v = elems[i];
				r = sndv(v);
				if (Verbose > 0)
					dbg_print("	t_ack: i=%d, v=%d, sndv=%d\n", i, v, r);
				if (r <= 0)
					goto error;
			}

			n = 0;
			p = rm_q(&free, &used, elems, &n);
			v = -free;
			r = sndv(v);
			if (Verbose > 0)
				dbg_print("	t_ack: free=%d, sndv=%d, stop=%d\n",
					free, r, stop);
			if (r <= 0)
				goto error;
			if (stop) return 0;
		} while (p > 20);
	}

 error:
	return 0;
}

static void
testev(uint sz, int adds, int rms)
{
	int ret;
	t_ctx_T tctx1, tctx2;

	ret = t_init(sz);
	if (ret != 0)
		return;

	tctx1.t_called = 0;
	tctx1.t_once = adds;
	ret = pthread_create(&task1, NULL, t_rcv, (void *) &tctx1);
	SM_TEST(ret == 0);
	SM_TEST(task1 != NULL);

	tctx2.t_called = 0;
	tctx2.t_once = rms;
	tctx2.t_max = 80;
	ret = pthread_create(&task2, NULL, t_ack, (void *) &tctx2);
	SM_TEST(ret == 0);
	SM_TEST(task2 != NULL);

	ret = pthread_join(task1, NULL);
	/* SM_TEST(ret == 0); */
	dbg_print("join1=%d\n", ret);

	ret = pthread_join(task2, NULL);
	/* SM_TEST(ret == 0); */
	dbg_print("join2=%d\n", ret);

	dbg_print("adds=%ld, rms=%ld\n", q_sent, q_rcvd);
}

static void
usage(const char *prg)
{
	sm_io_fprintf(smioerr, "%s: simple producer/consumer test\n", prg);
	sm_io_fprintf(smioerr, "usage %s [options]\n", prg);
	sm_io_fprintf(smioerr, "options\n");
	sm_io_fprintf(smioerr, "-a n  add n items on each iteration\n");
	sm_io_fprintf(smioerr, "-d n  remove n items on each iteration\n");
	sm_io_fprintf(smioerr, "-r n  run n loops\n");
	sm_io_fprintf(smioerr, "-s n  use n for size of window\n");
	sm_io_fprintf(smioerr, "-V    increase verbosity\n");
}

int
main(int argc, char *argv[])
{
	int c, adds, rms;
	uint sz;

	reps = 1000;
	sz = 100;
	adds = rms = 10;
	while ((c = getopt(argc, argv, "a:d:r:s:V")) != -1)
	{
		switch (c)
		{
		  case 'a':
			adds = atoi(optarg);
			break;
		  case 'd':
			rms = atoi(optarg);
			break;
		  case 'r':
			reps = atoi(optarg);
			break;
		  case 's':
			sz = atoi(optarg);
			break;
		  case 'V':
			Verbose++;
			break;
		  default:
			usage(argv[0]);
			return(1);
		}
	}
	testev(sz, adds, rms);
	return 0;
}
